#include <stdlib.h>
#include <stdio.h>
#include "basic.h"

static int verbose = 0 ;


typedef struct {
  int size ;
  volatile int *x,*y,*r0,*sync;
} common_t ;

typedef struct {
  int id ;
  common_t *common ;
} ctx_t ;

inline static void wait_partner(volatile int *p,int k,int id) {
  if (k % 2 == id) {
    *p = 1 ;
    __sync_synchronize() ;
  } else {
    while (*p == 0) ;
  }
}

void *T1(void *_p) {
  ctx_t *p = _p ;
  common_t *q = p->common ;
  int size = q->size ;
  for (int k = size-1 ; k >= 0 ; k--) {
    wait_partner(&q->sync[k],k,0) ;
    q->x[k] = 1 ;
    q->y[k] = 1 ;
  }
  return NULL ;
}

void *T2(void *_p) {
  ctx_t *p = _p ;
  common_t *q = p->common ;
  int size = q->size ;
  for (int k = size-1 ; k >= 0 ; k--) {
    wait_partner(&q->sync[k],k,1) ;
    q->y[k] = 2 ;
    int r0 = q->x[k] ;
    q->r0[k] = r0 ;
  }
  return NULL ;
}


typedef struct {
  int id ;
  int size ;
  int *count ;
  pthread_mutex_t *mutex ;
} run_t ;


void *run(void *_p) {
  run_t *p = (run_t *)_p ;
  int size = p->size ;

  int *x = calloc(size,sizeof(x[0])) ;
  int *y = calloc(size,sizeof(y[0])) ;
  int *r0 = calloc(size,sizeof(r0[0])) ;
  int *sync = calloc(size,sizeof(r0[0])) ;

  for ( ; ; ) {
    for (int k = 0 ; k < size ; k++) {
      x[k] = y[k] = 0 ; sync[k] = 0 ;
      r0[k] = -1 ;
    }
    common_t c ;
    c.size = size ; c.x = x ; c.y = y ; c.r0 = r0 ; c.sync = sync;
    pthread_t th1,th2 ;
    ctx_t a1,a2 ;
    
    a1.id = 1 ; a1.common = &c ;
    create_thread(&th1,T1,&a1) ;
    a2.id = 2 ; a2.common = &c ;
    create_thread(&th2,T2,&a2) ;
    join_thread(&th1) ;
    join_thread(&th2) ;

    for (int k = 0 ; k < size ; k++) INCR(&p->count[3*r0[k]+y[k]]) ;
    lock_mutex(p->mutex) ;

    printf("**** %i *****\n",p->id) ;
    int nz = 0 ;
    for (int k1 = 1; k1 < 3; k1++)
      for (int k2 = 0 ; k2 < 2 ; k2++) {
        int c = p->count[k2*3+k1] ;
        if (c > 0) nz++ ;
        printf("%-6i%c> y=%i; r0=%i\n",c,(k1 == 2 && k2 ==0 ? '*' : ':'),k1,k2) ;
      }
    unlock_mutex(p->mutex) ;
    if (nz == 4) break ;
  }
  free(x) ; free(y) ; free(r0) ; free(sync) ;
  return NULL ;
}

void zyva(int size, int nprocs) {
  int count[3*2] ;
  run_t args[nprocs] ;
  pthread_mutex_t *mtx = alloc_mutex() ;
  for (int i = 0 ; i < 3  ; i++)
    for (int j = 0 ; j < 2  ; j++)
      count[j*3+i] = 0 ;

  for (int id=0; id < nprocs; id++) {
    run_t *p = &args[id];
    p->count = &count[0] ;
    p->mutex = mtx;
    p->size = size ;
    p->id = id ;
  }

  pthread_t th[nprocs] ;
  for (int k = 0 ; k < nprocs ; k++) create_thread(&th[k],run,&args[k]) ;
  for (int k = 0 ; k < nprocs ; k++) join_thread(&th[k]) ;
  free_mutex(mtx) ;
}

void usage(char *pgm) {
  fprintf(stderr,"Usage: %s [-v]* [n]\n",pgm) ;
}

int main(int argc, char **argv) {
  char * pgm = argv[0] ;
  int size = SIZE ;
  int nprocs = 2 ;

  ++argv ;
  --argc ;
  while (argc > 0) {
    char *arg = *argv ;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') verbose++ ;
      else usage(pgm) ;
    } else {
      break ;
    }
    ++argv;
    --argc ;
  }
  if (argc == 1) {    
    size = atoi(argv[0]) ;
  } else if (argc >= 2) {
    size = atoi(argv[0]) ;
    nprocs = atoi(argv[1]) ;
  }
  zyva(size,nprocs) ;
  return 0 ;
}
